/****************************************************************************
****************************************************************************/

/*
This is a part of the LiteStep Shell Source code.

Copyright (C) 1997-2001 The LiteStep Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/****************************************************************************
****************************************************************************/


#include <algorithm>
#include "common.h"
#include <commctrl.h>
#include <shellapi.h>
#include "../current/lsapi/lsapi.h"
#include "../current/lsapi/macros.h"
#include "utilities.h"
#include "taskbar.h"
#include "bangs.h"

// undocumented SwitchToThisWindow API (maybe this should be in LSAPI?)
void (WINAPI *SwitchToThisWindow)(HWND, int);

const char aboutStr[] = "taskbar3.dll: 0.304 alpha 3 (jesus_mjjg)";

Taskbar *taskbar;

vector<string> NoFlashList;
vector<string> RemoveList;
vector<string> HideList;
vector<string> AddList;
vector<string> NoScrollList;

//
// utility functions
//

#ifdef TASKBAR_DEBUG
void Taskbar::Debug(){
	TaskbarButtonIterator it;
	char tmpTitle[BUFFER_SIZE], tmpClass[BUFFER_SIZE];

	for (it = buttons.begin(); it != buttons.end(); ++it)
	{
 	    GetWindowText((*it)->hTask , tmpTitle, BUFFER_SIZE);
 		GetClassName((*it)->hTask , tmpClass, BUFFER_SIZE);
		LSLogPrintf(0,"Taskbar.dll","0x%X :: %s :: %s", (*it)->hTask , tmpClass, tmpTitle);
	}
	LSLogPrintf(0,"Taskbar.dll","%d", taskbar->backColor);
}
#endif

void Taskbar::OnMove( int xP, int yP )
{
	taskbarX = xP;
	taskbarY = yP;
    SetWindowPosition();
}

void Taskbar::SwitchToButton( const char *args ){
	HWND current, prev, next;
	current = prev = next = NULL;
	TaskbarButtonIterator it ;

	current = lastActive;

	it = buttons.begin();

	while (it != buttons.end() && (*it)->hTask  != next){
		if((*it)->hTask  == current){
			++it;
			if(it != buttons.end())
				next = (*it)->hTask ;
			
		}
		else{
			prev = (*it)->hTask ;
			++it;
		}
	}

	if(!strcmp(args, "next") && next != NULL)
		SwitchToThisWindow( next, 1 );
	else if (!strcmp(args, "prev") && prev != NULL)
		SwitchToThisWindow( prev, 1 );
}

void Taskbar::ShowPart( const char *args , int i ){
	if(!strcmp(args, "next")){
		if( taskbar->buttons.size() > taskbar->maxButtonCount )
		if( taskbar->buttonCountStart + i <= taskbar->buttons.size() - taskbar->maxButtonCount )
			taskbar->buttonCountStart = taskbar->buttonCountStart + i;
	}
	else if(!strcmp(args, "prev"))
		if( taskbar->buttonCountStart - i >= 0 )
			taskbar->buttonCountStart = taskbar->buttonCountStart - i;
}

void Taskbar::GetRCFonts()
{
	NONCLIENTMETRICS ncm, ncm2;
	ncm2.cbSize = ncm.cbSize = sizeof(NONCLIENTMETRICS);

	SystemParametersInfo(SPI_GETNONCLIENTMETRICS, sizeof(NONCLIENTMETRICS), &ncm, 0);

	LOGFONT lf, lf2;
	LOGFONT *lfDefault = &ncm.lfCaptionFont;

	memset(&lf, 0, sizeof(LOGFONT));
	memset(&lf2, 0, sizeof(LOGFONT));

	//font settings
	GetRCString("TaskbarFont", lf.lfFaceName, lfDefault->lfFaceName, LF_FACESIZE);
	lf.lfHeight = GetRCInt("TaskbarFontHeight", lfDefault->lfHeight);
	lf.lfWeight = GetRCBool("TaskbarFontBold", TRUE) ? FW_BOLD : FW_NORMAL;
	lf.lfItalic = GetRCBool("TaskbarFontItalic", TRUE);
	lf.lfUnderline = GetRCBool("TaskbarFontUnderLine", TRUE);
    font = CreateFontIndirect(&lf);

	//active font settings
	GetRCString("TaskbarActiveFont", lf2.lfFaceName, lf.lfFaceName, LF_FACESIZE);
	lf2.lfHeight = GetRCInt("TaskbarActiveFontHeight", lf.lfHeight);
	lf2.lfWeight = GetRCBool("TaskbarActiveFontBold", TRUE) ? FW_BOLD : FW_NORMAL;
	lf2.lfItalic = GetRCBool("TaskbarActiveFontItalic", TRUE);
	lf2.lfUnderline = GetRCBool("TaskbarActiveFontUnderLine", TRUE);
    activeFont = CreateFontIndirect(&lf2);

	//minimized font settings
	GetRCString("TaskbarMinFont", lf2.lfFaceName, lf.lfFaceName, LF_FACESIZE);
	lf2.lfHeight = GetRCInt("TaskbarMinFontHeight", lf.lfHeight);
	lf2.lfWeight = GetRCBool("TaskbarMinFontBold", TRUE) ? FW_BOLD : FW_NORMAL;
	lf2.lfItalic = GetRCBool("TaskbarMinFontItalic", TRUE);
	lf2.lfUnderline = GetRCBool("TaskbarMinFontUnderLine", TRUE);
    minimizedFont = CreateFontIndirect(&lf2);

	//flashing font settings
	GetRCString("TaskbarFlashFont", lf2.lfFaceName, lf.lfFaceName, LF_FACESIZE);
	lf2.lfHeight = GetRCInt("TaskbarFlashFontHeight", lf.lfHeight);
	lf2.lfWeight = GetRCBool("TaskbarFlashFontBold", TRUE) ? FW_BOLD : FW_NORMAL;
	lf2.lfItalic = GetRCBool("TaskbarFlashFontItalic", TRUE);
	lf2.lfUnderline = GetRCBool("TaskbarFlashFontUnderLine", TRUE);
    flashFont = CreateFontIndirect(&lf2);

}



//
// Taskbar
//

Taskbar::Taskbar()
: hWnd(NULL), hInstance(NULL), hToolTips(NULL), edge(EDGE_BOTTOM), height(0), width(0),
  neededHeight(0), screenWidth(0), active(FALSE), mouseOver(FALSE),
  capture(NULL), skin(NULL), buttonSkin(NULL), activeButtonSkin(NULL), flashSkin(NULL),
  minimizedSkin(NULL), activeFont(0), font(0), minimizedFont(0), lastActive(GetActiveWindow()),
  taskbarLines(1), flashCount(0), scrollWas(GetKeyState(VK_SCROLL) != 0), resetScroll(false),  alwaysOnTop(FALSE),
  LsBoxed(false), timerScrollEnabled(false), reallyHidden(false), visible(false),
  buttonCountStart(0), useRealFlashing(false), taskbarTextShadow(false), textMinimizedNoShadow(FALSE),
  textNormalNoShadow(FALSE), textSelectedNoShadow(FALSE), textFlashingNoShadow(FALSE), switchOnHover(false), useToolTips(true)
{
	// ...
}

Taskbar::~Taskbar()
{
	// ...
}

UINT messages[] = {
                      LM_LSSELECT,
                      LM_GETMINRECT,
                      LM_REDRAW,
                      LM_WINDOWACTIVATED,
                      LM_WINDOWCREATED,
                      LM_WINDOWDESTROYED,
					  LM_GETREVID,
					  LM_REFRESH,
                      0
                  };

void Taskbar::OnLoad( HINSTANCE hInstanceP )
{
	hInstance = hInstanceP;

	// get config variables from step.rc
	ReadConfig();

	// register taskbar window class
	WNDCLASSEX wc;
	memset( &wc, 0, sizeof(WNDCLASSEX) );

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS;
	wc.lpfnWndProc = Taskbar::WndProc;
	wc.cbWndExtra = sizeof(Taskbar *);
	wc.hInstance = hInstance;
	wc.hCursor = LoadCursor( NULL, IDC_ARROW );
	wc.lpszClassName = WC_TASKBAR;

	if ( !RegisterClassEx( &wc ) )
	{
		RESOURCE_MSGBOX( hInstance, IDS_TASKBAR_ERROR1,
		                 "Could not register taskbar window class.", "Taskbar")
		return ;
	}

	// create the taskbar window
	 HWND parentWindow = NULL;
	 if(!alwaysOnTop){
		 parentWindow = FindWindow(WC_SHELLDESKTOP, NULL);
		 if (!parentWindow)
			 parentWindow = GetDesktopWindow();
	 }
     // create the taskbar window
     hWnd = CreateWindowEx( alwaysOnTop? WS_EX_TOOLWINDOW | WS_EX_TOPMOST : WS_EX_TOOLWINDOW, //666
                            WC_TASKBAR,
                            NULL,
                            WS_POPUP,
                            0, 0, 0, 0,
                            parentWindow,
                            NULL,
                            hInstance,
                            (LPVOID) this );
    //drag and drop support
	//DragAcceptFiles(hWnd, true);

	if ( !hWnd )
	{
		RESOURCE_MSGBOX( hInstance, IDS_TASKBAR_ERROR2,
		                 "Could not create taskbar window.", "Taskbar")

		return ;
	}

	// create tooltips control
	if(useToolTips)//20021108
	{
		hToolTips = CreateWindowEx( WS_EX_TOOLWINDOW | WS_EX_TOPMOST,
			                        TOOLTIPS_CLASS,
									NULL,
									TTS_ALWAYSTIP | WS_POPUP,
									0, 0, 0, 0,
									hWnd,
									NULL,
									hInstance,
									NULL );

		SetWindowPos( hToolTips, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE); //20021108
	}

	// adjust the Z-order, size, position
	SetWindowPos( hWnd, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER);
    
	// hide us from the VWM
	SetWindowLong( hWnd, GWL_USERDATA, magicDWord );

	// create font and compute needed height
	OnSettingChange();//should be in ReadConfig()
    SetWindowPosition();//removed from OnSettingChange to avoid taskbar resizing when changing wallpapers

	if( visible )
		ShowWindow(hWnd, SW_SHOWNOACTIVATE);

	// enumerate current tasks, setup buttons and start update timer
	EnumWindows(Taskbar::EnumWindowsProc, (LPARAM) this);
	Layout();

	SetTimer(hWnd, TIMER_UPDATE, timerRefresh, NULL);
	SetTimer(hWnd, TIMER_UPDATE_SCROLL, timerScroll, NULL);

	// register for LS messages
	SendMessage( GetLitestepWnd(), LM_REGISTERMESSAGE,
	             (WPARAM) hWnd, (LPARAM) ::messages );
}

void Taskbar::OnUnload()
{
	// shutdown timer
	KillTimer(hWnd, TIMER_UPDATE);
	KillTimer(hWnd, TIMER_UPDATE_SCROLL);

	// unregister messages
	SendMessage( GetLitestepWnd(), LM_UNREGISTERMESSAGE,
	             (WPARAM) hWnd, (LPARAM) ::messages );

	// destroy tooltips control
	if ( hToolTips )
		DestroyWindow( hToolTips );

	// destroy window and unregister class
	if ( hWnd )
		DestroyWindow( hWnd );

	UnregisterClass( WC_TASKBAR, hInstance );

	// release skins
	delete skin;
	delete buttonSkin;
	delete activeButtonSkin;
	delete minimizedSkin;
	delete flashSkin;

	// release fonts
	DeleteObject(activeFont);
	DeleteObject(font);
	DeleteObject(minimizedFont);
	DeleteObject(activeFont);

	// release buttons
	for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
	{
		delete (*it);
		*it = NULL;
	}

	buttons.clear();
}

void Taskbar::hide()
{
	if (visible)
	{
		visible = false;
		ShowWindow(hWnd, SW_HIDE);
	}
}

void Taskbar::show()
{
	if (!visible)
	{
		visible = true;
		ShowWindow(hWnd, SW_SHOWNOACTIVATE);
	}
}

void Taskbar::Invalidate( int x, int y, int cx, int cy )
{
	RECT r;

	r.left = x;
	r.top = y;
	r.right = x + cx;
	r.bottom = y + cy;

	InvalidateRect( hWnd, &r, FALSE );
}

TaskbarButton *Taskbar::GetCapture()
{
	return capture;
}

void Taskbar::SetCapture( TaskbarButton *button )
{
	::SetCapture( hWnd );
	capture = button;
}

void Taskbar::ReleaseCapture()
{
	::ReleaseCapture();
	capture = NULL;
}

TaskbarButton *Taskbar::ButtonAtPoint( int x, int y )
{
	for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
	{
		if ( (*it)->HasPoint(x, y) )
		{
			return (*it);
		}
	}

	return NULL;
}

int Taskbar::GetRCMouse(const char * rcString, int intDefault){
	char buffer[MAX_PATH];

	if ( GetRCString( rcString, buffer, "", 16 ) ){
		if ( !stricmp( buffer, ".select" ) )
			return MOUSE_SELECT;
		else if ( !stricmp( buffer, ".menu" ) )
			return MOUSE_MENU;
		else if ( !stricmp( buffer, ".close" ) )
			return MOUSE_CLOSE;
		else if ( !stricmp( buffer, ".close2" ) )
			return MOUSE_CLOSE2;
		else if ( !stricmp( buffer, ".ontop" ) )
			return MOUSE_ONTOP;
		else
			return MOUSE_SELECT;
	}
	else
		return intDefault;
}

void Taskbar::GetRCList(char *text, vector<string> &theList){
	//code from www.mindjunction.com
	FILE *f;
	f = LCOpen (NULL);
	if (f){
		char buffer[BUFFER_SIZE];
 		char token1[1204], token2[BUFFER_SIZE], extra_text[BUFFER_SIZE];
 		char* tokens[2];
	 
 		tokens[0] = token1;
 		tokens[1] = token2;
 
	 	buffer[0] = 0;

		while (LCReadNextConfig (f, text, buffer, sizeof (buffer)))
		{
			int count;
 
			token1[0] = token2[0] = extra_text[0];
  
 			count = LCTokenize (buffer, tokens, 2, extra_text);
  
  			// token 1 is "*TaskbarXXXX", token 2 is the
  			// application class or title, and the extra_text are something to forget.
  			if (count >= 2)
				theList.push_back(tokens[1]);
 		}
	}
}

void Taskbar::GetRCRect(char *text, int &left, int &right, int &top, int &bottom, int dv_left, int dv_right, int dv_top, int dv_bottom){
	char buffer[BUFFER_SIZE];
 	char token1[1204], token2[BUFFER_SIZE], token3[BUFFER_SIZE], token4[BUFFER_SIZE], extra_text[BUFFER_SIZE];
 	char* tokens[4];
	 
 	tokens[0] = token1;
 	tokens[1] = token2;
 	tokens[2] = token3;
 	tokens[3] = token4;
 
	buffer[0] = 0;

    if(GetRCLine(text, buffer, 1024, "")){
		int count;
 
		token1[0] = token2[0] = token3[0] = token4[0] = extra_text[0];
  
	 	count = LCTokenize (buffer, tokens, 4, extra_text);
        if(count >= 1){
			if(count >=4){
				left = atoi(token1);
				right = atoi(token2);
				top = atoi(token3);
				bottom = atoi(token4);
			}
			else{
				right = top = bottom = left = atoi(token1);
			}
		}
		else{
			right = dv_right;
			top = dv_top;
			bottom = dv_bottom;
			left = dv_left;
		}
	}
	else{
		right = dv_right;
		top = dv_top;
		bottom = dv_bottom;
		left = dv_left;
	}
}

void Taskbar::ReadConfig()
{
	char buffer[MAX_PATH];
	char bufferLeft[MAX_PATH];
	char bufferRight[MAX_PATH];

	// flash
	isTaskFlash = GetRCBool( "TaskButtonFlash", TRUE);
	visible = (!GetRCBool( "TaskbarStartHidden", TRUE));

	// edge
	if ( GetRCString( "TaskbarEdge", buffer, "", 16 ) )
	{
		edge = EDGE_BOTTOM;
		if ( !stricmp( buffer, "TOP" ) )
			edge = EDGE_TOP;
	}

	// appearance
	msTaskbar = GetRCBool( "TaskbarMSColors", TRUE );
	foreColor1 = GetRCColor( "TaskbarFore", RGB( 255, 255, 255 ) );
	foreColor2 = GetRCColor( "TaskbarFore2", RGB( 128, 128, 128 ) );
	backColor = GetRCColor( "TaskbarBack", RGB( 192, 192, 192 ) );
	activeColor = GetRCColor( "TaskbarActive", RGB( 192, 192, 192 ) );
	minColor = GetRCColor( "TaskbarMin", RGB( 192, 192, 192 ) );
	flashColor = GetRCColor( "TaskbarFlash", RGB( 224, 224, 224 ) );

	// skin settings
	noSkinShift = GetRCBool( "TaskbarNoSkinShift", TRUE );
	noTextShift = GetRCBool( "TaskbarNoTextShift", TRUE );

	// taskbar skin
	if ( GetRCString( "TaskbarSkin", buffer, "", MAX_PATH ) )
	{
		GetRCString( "TaskbarLeft", bufferLeft, "", MAX_PATH );
		GetRCString( "TaskbarRight", bufferRight, "", MAX_PATH );

		skin = new TaskbarSkin( bufferLeft, buffer, bufferRight );
	}

	// taskbar button skin
	if ( GetRCString( "TaskButtonSkin", buffer, "", MAX_PATH ) )
	{
		GetRCString( "TaskButtonLeft", bufferLeft, "", MAX_PATH );
		GetRCString( "TaskButtonRight", bufferRight, "", MAX_PATH );

		buttonSkin = new TaskbarSkin( bufferLeft, buffer, bufferRight );
	}

	// active taskbar button skin
	if ( GetRCString( "TaskButtonSkinActive", buffer, "", MAX_PATH ) )
	{
		GetRCString( "TaskButtonLeftActive", bufferLeft, "", MAX_PATH );
		GetRCString( "TaskButtonRightActive", bufferRight, "", MAX_PATH );

		activeButtonSkin = new TaskbarSkin( bufferLeft, buffer, bufferRight );
	}

	// minimized taskbar button skin
	if ( GetRCString( "TaskButtonSkinMin", buffer, "", MAX_PATH ) )
	{
		GetRCString( "TaskButtonLeftMin", bufferLeft, "", MAX_PATH );
		GetRCString( "TaskButtonRightMin", bufferRight, "", MAX_PATH );

		minimizedSkin = new TaskbarSkin( bufferLeft, buffer, bufferRight );
	}

	// taskbar button flash skin
	if ( GetRCString( "TaskButtonFlashSkin", buffer, "", MAX_PATH ) )
	{
		GetRCString( "TaskButtonFlashLeft", bufferLeft, "", MAX_PATH );
		GetRCString( "TaskButtonFlashRight", bufferRight, "", MAX_PATH );

		flashSkin = new TaskbarSkin( bufferLeft, buffer, bufferRight );
	}

	GetRCFonts();

	textColor = GetRCColor( "TaskbarText", msTaskbar ? GetSysColor( COLOR_BTNTEXT ) : RGB( 255, 255, 255 ) );
	textActiveColor = GetRCColor( "TaskbarTextActive", msTaskbar ? GetSysColor( COLOR_BTNTEXT ) : RGB( 192, 192, 192 ) );
    textMinimizedColor = GetRCColor( "TaskbarTextMin", msTaskbar ? GetSysColor( COLOR_BTNTEXT ) : RGB( 192, 192, 192 ) );
	textFlashColor = GetRCColor( "TaskbarTextFlash", msTaskbar ? GetSysColor( COLOR_BTNTEXT ) : RGB( 64, 64, 64 ) );

	taskbarTextShadow = GetRCBool("TaskbarUseTextShadow", TRUE); 
	if(taskbarTextShadow){
	  COLORREF textShadowColor;
	  textShadowColor = GetRCColor( "TaskbarTextShadow", msTaskbar ? GetSysColor( COLOR_BTNTEXT ) : RGB( 255, 255, 255 ) );

	  textMinimizedShadowColor = GetRCColor( "TaskbarMinimizedShadowColor", textShadowColor);
	  textSelectedShadowColor = GetRCColor( "TaskbarSelectedShadowColor", textShadowColor);
	  textNormalShadowColor = GetRCColor( "TaskbarNormalShadowColor", textShadowColor);
	  textFlashingShadowColor = GetRCColor( "TaskbarFlashingShadowColor", textShadowColor);

	  textMinimizedNoShadow = GetRCBool("TaskbarMinimizedNoShadow", TRUE);
	  textNormalNoShadow = GetRCBool("TaskbarNormalNoShadow", TRUE);
	  textSelectedNoShadow = GetRCBool("TaskbarSelectedNoShadow", TRUE);
	  textFlashingNoShadow = GetRCBool("TaskbarFlashingNoShadow", TRUE);


  	  textShadowOffsetX = GetRCInt( TEXT("TaskBarShadowOffsetX"), 1);
	  textShadowOffsetY = GetRCInt( TEXT("TaskBarShadowOffsetY"), 1);
	}

	
	taskbarPositioning = GetRCBoolDef("TaskbarPositioning", FALSE);
	if (taskbarPositioning)
	{
		taskbarX = GetRCInt("TaskbarX", 0);
		taskbarXS = taskbarX;
		if (taskbarX < 0)
			taskbarX = SCREEN_WIDTH + taskbarX;
		taskbarY = GetRCInt("TaskbarY", 0);
		taskbarYS = taskbarY;
		if (taskbarY < 0)
			taskbarY = SCREEN_HEIGHT + taskbarY;
	}

	// SEE: OnSettingChange for width setting TaskbarWidth

	alwaysOnTop = GetRCBool("TaskbarAlwaysOnTop", TRUE);
	reallyHidden = GetRCBool("TaskbarHideTasks", TRUE);
	
	taskbarHeight = GetRCInt( TEXT("TaskBarHeight"), 0 );

	screenWidth = GetRCInt( TEXT("TaskBarWidth"), SCREEN_WIDTH );
	
	maxTaskWidth = GetRCInt( TEXT("TaskbarMaxTaskWidth"), SCREEN_WIDTH );

	GetRCRect("TaskbarBorder", borderLeft, borderRight, borderTop, borderBottom, 4, 4, 4, 4);
	GetRCRect("TaskbarButtonBorder", buttonBorderLeft, buttonBorderRight, buttonBorderTop, buttonBorderBottom, 4, 4, 4, 4);
	GetRCRect("TaskbarButtonTextBorder", buttonTextBorderLeft, buttonTextBorderRight, buttonTextBorderTop, buttonTextBorderBottom, 20, 0, 0, 0);

	spacing = GetRCInt( TEXT("TaskBarSpacing"), 3 );
	iconSize = GetRCInt( TEXT("TaskBarIconSize"), 16);

	showIcon = GetRCBoolDef("TaskbarIconShow", TRUE) ? true : false;

    /*mouse buttons on bar*/
	GetRCLine("TaskBarLButtonDown", taskbarLButtonDown, MAX_LINE_LENGTH, "!none");
	GetRCLine("TaskBarRButtonDown", taskbarRButtonDown, MAX_LINE_LENGTH, "!none");
	GetRCLine("TaskBarMButtonDown", taskbarMButtonDown, MAX_LINE_LENGTH, "!none");
	GetRCLine("TaskBarLButtonUp", taskbarLButtonUp, MAX_LINE_LENGTH, "!none");
	GetRCLine("TaskBarRButtonUp", taskbarRButtonUp, MAX_LINE_LENGTH, "!none");
	GetRCLine("TaskBarMButtonUp", taskbarMButtonUp, MAX_LINE_LENGTH, "!none");

	/*action on tiles*/
	TaskBarTileLButton = GetRCMouse("TaskBarTButtonL", MOUSE_SELECT);
	TaskBarTileMButton = GetRCMouse("TaskBarTButtonM", MOUSE_NONE);
	TaskBarTileRButton = GetRCMouse("TaskBarTButtonR", MOUSE_MENU);

	taskbarLines =  GetRCInt( TEXT("TaskBarLines"), 1 );
	if (taskbarLines <= 0)
	  taskbarLines = 1;

	timerRefresh =  GetRCInt( TEXT("TaskBarRefresh"), 500 );
    if (timerRefresh < 100)
		timerRefresh = 500;
	timerScroll =  GetRCInt( TEXT("TaskBarScrollFlash"), 500 );
    if (timerScroll < 100)
		timerScroll = 500;
	timerScrollEnabled = GetRCBool( "TaskbarFlashScrollLock", TRUE );

	//Taskbar Lists
    /*vector<string> HideList;*/
	GetRCList("*TaskbarNoFlash", NoFlashList);
	GetRCList("*TaskbarRemove", RemoveList);
	GetRCList("*TaskbarAdd", AddList);
	GetRCList("*TaskbarNoScroll",NoScrollList);
	GetRCList("*TaskbarHide", HideList);

	iconX = GetRCInt( TEXT("TaskBarIconX"), 0);
	iconY = GetRCInt( TEXT("TaskBarIconY"), 0);

	if ( GetRCString( "TaskbarHorizontalTextAlign", buffer, "", 16 ) )
	{
		if ( !stricmp( buffer, ".center" ) )
			textX = -2;
		else if ( !stricmp( buffer, ".right" ) )
			textX = -3;
		else
			textX = -1;
	}
	else
		textX = -1;

	if ( GetRCString( "TaskbarVerticalTextAlign", buffer, "", 16 ) )
	{
		if ( !stricmp( buffer, ".center" ) )
			textY = -2;
		else if ( !stricmp( buffer, ".bottom" ) )
			textY = -3;
		else
			textY = -1;
	}
	else
		textY = -2;

	showText = GetRCBoolDef("TaskbarTextShow", TRUE) ? true : false;

	if ( GetRCString( "TaskbarButtonDirection", buffer, "", 16 ) )
	{
		if ( !stricmp( buffer, ".left" ) )
			direction = DIRECTION_LEFT;
		else
			direction = DIRECTION_RIGHT;
	}
	else
		direction = DIRECTION_RIGHT;

	maxButtonCount	= GetRCInt("TaskbarMaxButtonCount", SCREEN_WIDTH);
	if( maxButtonCount <= 0)
		maxButtonCount = SCREEN_WIDTH;
	useRealFlashing = GetRCBool("TaskbarUseRealFlashing", TRUE);
	//20021108
	switchOnHover = GetRCBool("TaskbarSwitchOnMouseHover", TRUE);
	//20021110
	if (GetRCString( "TaskbarUseToolTips", buffer, "", 16 ))
	{
		if ( !stricmp( buffer, "false" ) )
			useToolTips = false;
	}
	//useToolTips = GetRCBool("TaskbarUseToolTips", TRUE);
	minimizeOnHover = GetRCBool("TaskbarMinimizeOnMouseHover", TRUE);
}

void Taskbar::Resize(int x, int y){
	MoveWindow( taskbar->hWnd ,  taskbar->taskbarX, taskbar->taskbarY, taskbar->width + x,taskbar->height + y, TRUE);
}

void Taskbar::Layout()
{
	inCriticalProcess = true;
	int buttonCount = buttons.size();
	if( taskbar->reallyHidden ){
		for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
			if ( IsOnTaskbarList((*it)->hTask , HideList) )
				--buttonCount;
/*			if(!IsAppWindow((*it)->hTask )){
				TaskbarButton *button = *it;
				delete button;
				buttons.erase( it );

				if(taskbar->buttonCountStart>0)
					taskbar->buttonCountStart -=1;
			}*/
	}

    //on ne doit montrer qu'une partie des boutons
	if ( buttonCount > maxButtonCount)
	    buttonCount = maxButtonCount;

	if (buttonCount <= 0)
	{
		InvalidateRect( hWnd, NULL, FALSE );
		inCriticalProcess = false;
		return ;
	}

	int availableWidth = width - borderLeft - borderRight;
	availableWidth = max( availableWidth, 0 );  // cannot be less than zero

	if ( buttonCount > 1 )
	{
		availableWidth = availableWidth - (buttonCount - 1) * spacing;
		availableWidth = max( availableWidth, 0 );
	}

	int widthPerButton = min(availableWidth / buttonCount,maxTaskWidth);

	int buttonHeight = (height - borderTop - borderBottom - (taskbarLines - 1) * spacing) / taskbarLines;

	int lines =0;
	int linebuttonCount;

    if(taskbarLines>1){
		if (!(buttonCount%taskbarLines))
			linebuttonCount = ((buttonCount) / taskbarLines);
		else
			linebuttonCount = ((buttonCount) / taskbarLines) + 1;
	}
	else
	  linebuttonCount = buttonCount;

    int error = availableWidth % linebuttonCount;
    int errorDone = 0;
    widthPerButton = min(availableWidth / linebuttonCount,maxTaskWidth);

	//vers la droite
	int x;
	if(taskbar->direction == DIRECTION_RIGHT)
		x = borderLeft;
	else
		x = taskbar->width - borderRight - widthPerButton;

	int y = borderTop;

	int buttonNum = 0;
	// set placement for each button
	for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
	{
		TaskbarButton *button = *it;

		if((taskbar->reallyHidden && !IsOnTaskbarList((*it)->hTask , HideList)) || !taskbar->reallyHidden){
			if( ( buttonNum >= buttonCountStart ) && ( buttonNum - buttonCountStart < maxButtonCount ) ){
				if ( error > 0 )
				{
					++lines;
					if((lines % taskbarLines) == 0){
			  			error--;
				  		errorDone++;
					}
 		    		button->Move( x , y, widthPerButton + 1, buttonHeight );

					if(taskbar->direction == DIRECTION_RIGHT)
						x = borderLeft + (lines / taskbarLines) * widthPerButton + (lines / taskbarLines) * spacing + errorDone;
					else 
						x = taskbar->width - borderRight - (lines / taskbarLines + 1 ) * widthPerButton - (lines / taskbarLines) * spacing - errorDone;
	            	y = borderTop + (lines % taskbarLines) * buttonHeight + (lines % taskbarLines) * spacing;
				}
				else
				{
					button->Move( x , y, widthPerButton, buttonHeight );
					++lines;

					if(taskbar->direction == DIRECTION_RIGHT)
						x = borderLeft + (lines / taskbarLines) * widthPerButton + (lines / taskbarLines) * spacing + errorDone;
					else
						x = taskbar->width - borderRight - (lines / taskbarLines + 1) * widthPerButton - (lines / taskbarLines) * spacing - errorDone;
					y = borderTop + (lines % taskbarLines) * buttonHeight + (lines % taskbarLines) * spacing;
				}
			}
			else
				button->Move( x , y, 0, 0);
			++buttonNum;
		}
		else
			button->Move( x , y, 0, 0);
	}

	InvalidateRect( hWnd, NULL, FALSE );
	inCriticalProcess = false;
}



void Taskbar::SetWindowPosition()
{
	if (taskbarPositioning)
	{
		SetWindowPos( hWnd,  NULL, taskbarX, taskbarY,screenWidth ,neededHeight, SWP_NOACTIVATE | SWP_NOZORDER );
	}
	else
	{
		if ( edge == EDGE_TOP )
		{
			SetWindowPos( hWnd, NULL, 0, 0, screenWidth, neededHeight, SWP_NOACTIVATE | SWP_NOZORDER );
		}
		else
		{
			SetWindowPos( hWnd, NULL, 0, SCREEN_HEIGHT - neededHeight, screenWidth, neededHeight, SWP_NOACTIVATE | SWP_NOZORDER );
		}
	}
}

TaskbarButton *Taskbar::SearchForButtonByHandle( HWND hWnd )
{
	for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
	{
		if ( hWnd == (*it)->hTask  )
			return (*it);
	}

	return NULL;
}

void Taskbar::OnActivate( BOOL becomingActive )
{
	//what's the purpose of active?
	active = becomingActive;
}

void Taskbar::OnDisplayChange()
{
  // ....
}

void Taskbar::OnGetMinRect( HWND hWnd, RECT *rectangle )//CPU 20020501
{
	TaskbarButton *button;
	button = SearchForButtonByHandle(hWnd);
	//20021110
	//button->OnUpdate();
}

void Taskbar::OnLButtonDown( int x, int y )
{
	TaskbarButton *button;

	if ( capture )
		button = capture;
	else
		button = ButtonAtPoint( x, y );

	if ( button )
	{
		button->OnLButtonDown( x, y );
	}
	else{
		LSExecute(NULL, Taskbar::taskbarLButtonDown, NULL);
	}
}

void Taskbar::OnMButtonDown( int x, int y )
{
	TaskbarButton *button;

	if ( capture )
		button = capture;
	else
		button = ButtonAtPoint( x, y );

	if ( button )
	{
		button->OnMButtonDown( x, y );
	}
	else{
		LSExecute(NULL, Taskbar::taskbarMButtonDown, NULL);
	}
}

void Taskbar::OnMButtonUp( int x, int y )
{
	TaskbarButton *button;

	if ( capture )
		button = capture;
	else
		button = ButtonAtPoint( x, y );

	if ( button )
		button->OnMButtonUp( x, y );
	else
	{
		LSExecute(NULL, Taskbar::taskbarMButtonUp, NULL);
	}

}


void Taskbar::OnLButtonUp( int x, int y )
{
	TaskbarButton *button;

	if ( capture )
		button = capture;
	else
		button = ButtonAtPoint( x, y );

	if ( button )
		button->OnLButtonUp( x, y );
	else{
		if( taskbar->lastActive ){
			taskbar->lastActive = NULL;
		}
		LSExecute(NULL, Taskbar::taskbarLButtonUp, NULL);
	}
}

void Taskbar::OnMouseMove( int x, int y )
{
	mouseOver = TRUE;

	TaskbarButton *button;

	if ( capture )
		button = capture;
	else
		button = ButtonAtPoint( x, y );

	if ( button )
	{
		button->OnMouseMove( x, y );
		return ;
	}

	POINT pt;
	RECT r;

	pt.x = x;
	pt.y = y;

	r.left = 2;
	r.top = 4;
	r.right = r.left + 64;
	r.bottom = height - 4;

}

void Taskbar::OnPaint( HDC hDC )
{
	HDC hdcBuffer;
	HBITMAP hbmBuffer;
	HBITMAP hbmTemp;

	hdcBuffer = CreateCompatibleDC( hDC );
	hbmBuffer = CreateCompatibleBitmap( hDC, width, height );
	hbmTemp = (HBITMAP) SelectObject( hdcBuffer, hbmBuffer );

	if ( !skin )
	{
		RECT r;
		GetClientRect( hWnd, &r );

		if ( !msTaskbar )
		{
			// simple 3D frame
			HBRUSH hbrBack;
			HPEN hpnFore;
			HPEN hpnTemp;

			hbrBack = CreateSolidBrush( backColor );
			FillRect( hdcBuffer, &r, hbrBack );
			DeleteObject( hbrBack );

			hpnFore = CreatePen( PS_SOLID, 1, foreColor2 );
			hpnTemp = (HPEN) SelectObject( hdcBuffer, hpnFore );

			MoveToEx( hdcBuffer, r.left, r.bottom - 1, NULL );
			LineTo( hdcBuffer, r.right - 1, r.bottom - 1 );
			LineTo( hdcBuffer, r.right - 1, r.top );

			SelectObject( hdcBuffer, hpnTemp );
			DeleteObject( hpnFore );

			hpnFore = CreatePen( PS_SOLID, 1, foreColor1 );
			hpnTemp = (HPEN) SelectObject( hdcBuffer, hpnFore );

			LineTo( hdcBuffer, r.left, r.top );
			LineTo( hdcBuffer, r.left, r.bottom - 1 );

			SelectObject( hdcBuffer, hpnTemp );
			DeleteObject( hpnFore );
		}
		else
		{
			// MS taskbar style
			DrawEdge( hdcBuffer, &r, EDGE_RAISED, BF_RECT | BF_MIDDLE );
		}
	}
	else
	{
		// skin
		skin->Apply( hdcBuffer, 0, 0, width, height );
	}

	int buttonNum = 0;
	for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it){
		if( ( buttonNum >= buttonCountStart ) && ( buttonNum - buttonCountStart < maxButtonCount ) ){
			if ((taskbar->reallyHidden ) && (!IsOnTaskbarList((*it)->hTask , HideList)))
				(*it)->OnPaint( hdcBuffer );
			else if (!taskbar->reallyHidden )
				(*it)->OnPaint( hdcBuffer );
		}
		++buttonNum;
	}

	BitBlt( hDC, 0, 0, width, height, hdcBuffer, 0, 0, SRCCOPY );

	SelectObject( hdcBuffer, hbmTemp );
	DeleteObject( hbmBuffer );
	DeleteDC( hdcBuffer );
}

// OnRedraw
//

void Taskbar::OnRedraw( HWND hWnd, BOOL fFlash )
{
	TaskbarButton *button = SearchForButtonByHandle(hWnd);

	if (button)
	{
		button->SetFlashing(fFlash);
		if(useRealFlashing){
#ifdef TASKBAR_DEBUG
			LSLogPrintf(0,"Taskbar.dll","Flashing of 0x%X :: %d", hWnd , fFlash);
#endif
			if(fFlash)
				button->flashingFlashSkin = !(button->flashingFlashSkin) ; //200206111
			else
				button->flashingFlashSkin = true;
		}
		button->OnUpdate(TRUE);
	}
}

void Taskbar::OnRButtonDown( int x, int y )
{
	TaskbarButton *button;

	if ( capture )
		button = capture;
	else
		button = ButtonAtPoint( x, y );

	if ( button )
		button->OnRButtonDown( x, y );
	else
      LSExecute(NULL, this->taskbarRButtonDown, NULL);

}

void Taskbar::OnRButtonUp( int x, int y )
{
	TaskbarButton *button;

	if ( capture )
		button = capture;
	else
		button = ButtonAtPoint( x, y );

	if ( button )
		button->OnRButtonUp( x, y );
	else
      LSExecute(NULL, this->taskbarRButtonUp, NULL);
}

void Taskbar::OnSettingChange()
{
	// compute needed height and set initial position
	TEXTMETRIC tm;
	HDC hDC;
	HFONT hfnTemp;

	hDC = GetDC( NULL );
	hfnTemp = (HFONT) SelectObject( hDC, font );

	GetTextMetrics( hDC, &tm );

	SelectObject( hDC, hfnTemp );
	ReleaseDC( NULL, hDC );

    
	if(taskbarHeight <= 0){
	  neededHeight = max( (const int)(tm.tmHeight + tm.tmExternalLeading), iconSize );
  	  neededHeight = neededHeight + buttonBorderTop + buttonBorderBottom;
	  neededHeight = neededHeight + borderTop + borderBottom;
	}
	else
      neededHeight = taskbarHeight;
}

void Taskbar::OnSize( int widthP, int heightP )
{
	height = heightP;
	width = widthP;

	Layout();
}


void Taskbar::OnTimerUpdate(int whichTimer)
{
	//win95 buttons not disappearing
	/*for (TaskbarButtonIterator i = buttons.begin(); i != buttons.end(); ++i )
	{
		TaskbarButton *button = *i;

		if ( !IsAppWindow(button->hTask ) )
		{
			delete button;
			buttons.erase( i );

			if(taskbar->buttonCountStart>0)
				taskbar->buttonCountStart -=1;

			Layout();
			return ;
		}
	}*///win95 doesn't like timers ?

	TaskbarButtonIterator it;
    taskbar->flashCount = 0;
	for (it = buttons.begin(); it != buttons.end(); ++it){
		if ((*it)->getCurrentState(true) == STATE_FLASHING)
			if( !IsOnTaskbarList((*it)->hTask , NoScrollList) )
				taskbar->flashCount = taskbar->flashCount + 1;
	}

	if((whichTimer == TIMER_UPDATE_SCROLL)){
		if((taskbar->timerScrollEnabled)){
            if(taskbar->flashCount > 0){
				keybd_event( VK_SCROLL, 0x45, KEYEVENTF_EXTENDEDKEY | 0, 0 );
				keybd_event( VK_SCROLL, 0x45, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
			}
			/*else if ((scrollWas && !GetKeyState(VK_SCROLL)) || (!scrollWas && GetKeyState(VK_SCROLL))){
		  		keybd_event( VK_SCROLL, 0x45, KEYEVENTF_EXTENDEDKEY | 0, 0 );
				keybd_event( VK_SCROLL, 0x45, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
			}*/

		}
		else if (resetScroll == true){
			if ((scrollWas && !GetKeyState(VK_SCROLL)) || (!scrollWas && GetKeyState(VK_SCROLL))){
		  		keybd_event( VK_SCROLL, 0x45, KEYEVENTF_EXTENDEDKEY | 0, 0 );
				keybd_event( VK_SCROLL, 0x45, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
			}
			resetScroll = false;
		}
	}
	else{
		if(!inCriticalProcess)
			for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
			{
				(*it)->OnUpdate((*it)->getCurrentState(true) != (*it)->currentState);
				//(*it)->OnUpdate(((*it)->getCurrentState(true) != (*it)->currentState) || ((*it)->getCurrentState(true) == STATE_FLASHING));
			}
	}
}

void Taskbar::OnWindowActivated( HWND hApplication, BOOL fFullScreen ) //win95?
{
/*	int count = 0;
	TaskbarButtonIterator it;
	it = buttons.begin();
	while ((it != buttons.end()) && ((*it)->hTask  != hApplication)){
		//helps !multiple windows selecting ... acdsee 4
		if((*it)->hTask  == taskbar->lastActive)
			(*it)->OnUpdate(true);
		//helps !multiple windows selecting ... acdsee 4
		++it;
		++count;
	}
	if((count >= taskbar->buttonCountStart) && (count <= taskbar->buttonCountStart + taskbar->maxButtonCount));
	else{
		if(count < taskbar->buttonCountStart)
			taskbar->buttonCountStart = count;
		else
			taskbar->buttonCountStart = min(count,	taskbar->buttons.size() - taskbar->maxButtonCount);
//jesus_mjjg all tasks selected
		//taskbar->Layout();
	}*/
	TaskbarButton *button = SearchForButtonByHandle(hWnd);

	if (button)
	{
		button->flashingFlashSkin = false;
    }	
	taskbar->Layout();
//end jesus_mjjg all tasks selected
}

HWND TaskbarButton::buttonHandler(){
	return hTask;
}

/*BOOL Taskbar::IsAlreadyInList(HWND hWnd){
	for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
	{
	  if (hWnd == (*it)->buttonHandler())
		  return TRUE;
	}
    return FALSE;
}*/

void Taskbar::OnRefresh(  ){

	KillTimer(hWnd, TIMER_UPDATE);
	KillTimer(hWnd, TIMER_UPDATE_SCROLL);
	// release buttons
	for (TaskbarButtonIterator it = buttons.begin(); it != buttons.end(); ++it)
	{
		delete (*it);
		*it = NULL;
	}

	buttons.clear();

	ReadConfig();
	SetWindowPos( hWnd, alwaysOnTop? HWND_TOPMOST : HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	OnSettingChange();//should be in ReadConfig()
	SetWindowPosition();//removed from OnSettingChange to avoid taskbar resizing when changing wallpapers
	ShowWindow(hWnd, visible ? SW_SHOWNOACTIVATE : SW_HIDE);

	EnumWindows(Taskbar::EnumWindowsProc, (LPARAM) this);
	Layout();

	SetTimer(hWnd, TIMER_UPDATE, timerRefresh, NULL);
	SetTimer(hWnd, TIMER_UPDATE_SCROLL, timerScroll, NULL);
	// register for LS messages
	SendMessage( GetLitestepWnd(), LM_REGISTERMESSAGE,
	             (WPARAM) hWnd, (LPARAM) ::messages );
}

void Taskbar::OnWindowCreated( HWND hTask )
{
    if (SearchForButtonByHandle(hTask)==NULL && !IsOnTaskbarList(hTask, RemoveList)){
		TaskbarButton *button = new TaskbarButton( *this, hTask );
		button->SetSkin(buttonSkin);
		button->SetActiveSkin(activeButtonSkin);
		button->SetFlashSkin(flashSkin);
		button->SetMinimizedSkin(minimizedSkin);

		buttons.push_back( button );

	}
	Layout();
}

void Taskbar::OnWindowDestroyed( HWND hTask )
{
	for (TaskbarButtonIterator i = buttons.begin(); i != buttons.end(); ++i )
	{
		TaskbarButton *button = *i;

		if ( button->hTask  == hTask )
		{
			delete button;
			buttons.erase( i );

			if(taskbar->buttonCountStart>0)
				taskbar->buttonCountStart -=1;

//			Layout();
			return ;
		}
	}
}

BOOL Taskbar::OnWindowMessage( UINT nMessage, WPARAM wParam, LPARAM lParam, LRESULT &lResult )
{
	lResult = 0;

	// relay mouse messages to tooltips control
	if(useToolTips)//20021108
	if ( nMessage >= WM_MOUSEFIRST && nMessage <= WM_MOUSELAST )
	{
		MSG msg;

		msg.hwnd = hWnd;
		msg.message = nMessage;
		msg.wParam = wParam;
		msg.lParam = lParam;

		SendMessage( hToolTips, TTM_RELAYEVENT, 0, (LPARAM) &msg );
		SetWindowPos( hToolTips, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
	}

	switch ( nMessage )
	{
		case LM_LSSELECT:
		{
			OnWindowActivated( (HWND) wParam, FALSE );
			return TRUE;
		}

		case LM_GETMINRECT:
		{
			//OnGetMinRect( (HWND) wParam, (RECT *) lParam );
			LSLogPrintf(0,"taskbar3", "LM_GETMINRECT: %d", (HWND) wParam);
			return TRUE;
		}

		case LM_REDRAW:
		{
			OnRedraw( (HWND) wParam, (BOOL) lParam );
			return TRUE;
		}

		case LM_WINDOWACTIVATED:
		{
			OnWindowActivated( (HWND) wParam, (BOOL) lParam );
			return TRUE;
		}

		case LM_WINDOWCREATED:
		{
			OnWindowCreated( (HWND) wParam );
			return TRUE;
		}

		case LM_WINDOWDESTROYED:
		{
			OnWindowDestroyed( (HWND) wParam );
			Layout();
			return TRUE;
		}

		case LM_REFRESH:
		{
			OnRefresh();
			return TRUE;
		}

		case LM_GETREVID:
		{
			char* buf = (char*)lParam;
			strcpy(buf, aboutStr);
			return strlen(buf);
		}

		case WM_ACTIVATE:
		{
			OnActivate( (LOWORD(wParam) == WA_INACTIVE) ? FALSE : TRUE );
			return TRUE;
		}

		case WM_DISPLAYCHANGE:
		{
			OnDisplayChange();
			return TRUE;
		}

		case WM_LBUTTONDOWN:
		{
			OnLButtonDown( LOWORD(lParam), HIWORD(lParam));
			return TRUE;
		}
		case WM_LBUTTONUP:
		{
			OnLButtonUp( LOWORD(lParam), HIWORD(lParam));
			return TRUE;
		}
		case WM_MBUTTONDOWN:
		{
			OnMButtonDown( LOWORD(lParam), HIWORD(lParam));
			return TRUE;
		}
		case WM_MOUSEMOVE:
		{
			OnMouseMove( LOWORD( lParam ), HIWORD( lParam ) );
			return TRUE;
		}

		case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hDC;

			if ( !wParam )
				hDC = BeginPaint( hWnd, &ps );
			else
				hDC = (HDC) wParam;

			OnPaint( hDC );

			if ( !wParam )
				EndPaint( hWnd, &ps );

			return TRUE;
		}

		case WM_RBUTTONDOWN:
		{
			OnRButtonDown( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
			return TRUE;
		}

		case WM_RBUTTONUP:
		{
			OnRButtonUp( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
			return TRUE;
		}
		case WM_SIZE:
		{
			OnSize( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
			return TRUE;
		}

		case WM_TIMER:
		{
			switch ( wParam )
			{
				/*case TIMER_AUTOHIDE:
				{
					OnTimerAutoHide();
					return TRUE;
				}*/
				
				case TIMER_UPDATE:
				{
					OnTimerUpdate(TIMER_UPDATE);
					return TRUE;
				}
				case TIMER_UPDATE_SCROLL:
				{
					OnTimerUpdate(TIMER_UPDATE_SCROLL);
					return TRUE;
				}
			}

			return FALSE;
		}
	}

	return FALSE;
}

BOOL WINAPI Taskbar::EnumWindowsProc(HWND hWnd, LPARAM lParam)
{
	Taskbar *taskbar = (Taskbar *) lParam;
	if (taskbar && IsAppWindow(hWnd))
	{
		TaskbarButton *button = taskbar->SearchForButtonByHandle(hWnd);

		if (button == 0)
		{
			button = new TaskbarButton(*taskbar, hWnd);

			button->SetActiveSkin(taskbar->activeButtonSkin);
			button->SetSkin(taskbar->buttonSkin);
			button->SetFlashSkin(taskbar->flashSkin);
			button->SetMinimizedSkin(taskbar->minimizedSkin);

			taskbar->buttons.push_back(button);
			taskbar->listChanged = true;
		}
	}
	return TRUE;
}

LRESULT WINAPI Taskbar::WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	Taskbar *taskbar = (Taskbar *) GetWindowLong(hWnd, 0);

	if (message == WM_NCCREATE)
	{
		LPCREATESTRUCT pcs = (LPCREATESTRUCT) lParam;
		taskbar = (Taskbar *) pcs->lpCreateParams;

		taskbar->hWnd = hWnd;
		SetWindowLong(hWnd, 0, (LONG) taskbar);
	}

	if (taskbar)
	{
		LRESULT lResult;

		if (taskbar->OnWindowMessage(message, wParam, lParam, lResult))
			return lResult;
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}
